@@ -1,6 +1,6 @@
 /*
   SDL_image:  An example image loading library for use with SDL
-  Copyright (C) 1997-2018 Sam Lantinga <slouken@libsdl.org>
+  Copyright (C) 1997-2022 Sam Lantinga <slouken@libsdl.org>
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -26,6 +26,10 @@
 
 #ifdef LOAD_XCF
 
+#ifndef SDL_SIZE_MAX
+#define SDL_SIZE_MAX ((size_t)-1)
+#endif
+
 #if DEBUG
 static char prop_names [][30] = {
   "end",
@@ -138,9 +142,11 @@ typedef struct {
 
 typedef struct {
   char   sign [14];
+  Uint32 file_version;
   Uint32 width;
   Uint32 height;
   Sint32 image_type;
+  Uint32 precision;
   xcf_prop * properties;
 
   Uint32 * layer_file_offsets;
@@ -158,8 +164,8 @@ typedef struct {
   char * name;
   xcf_prop * properties;
 
-  Uint32 hierarchy_file_offset;
-  Uint32 layer_mask_offset;
+  Uint64 hierarchy_file_offset;
+  Uint64 layer_mask_offset;
 
   Uint32 offset_x;
   Uint32 offset_y;
@@ -172,7 +178,7 @@ typedef struct {
   char * name;
   xcf_prop * properties;
 
-  Uint32 hierarchy_file_offset;
+  Uint64 hierarchy_file_offset;
 
   Uint32 color;
   Uint32 opacity;
@@ -185,19 +191,19 @@ typedef struct {
   Uint32 height;
   Uint32 bpp;
 
-  Uint32 * level_file_offsets;
+  Uint64 * level_file_offsets;
 } xcf_hierarchy;
 
 typedef struct {
   Uint32 width;
   Uint32 height;
 
-  Uint32 * tile_file_offsets;
+  Uint64 * tile_file_offsets;
 } xcf_level;
 
 typedef unsigned char * xcf_tile;
 
-typedef unsigned char * (* load_tile_type) (SDL_RWops *, Uint32, int, int, int);
+typedef unsigned char * (*load_tile_type) (SDL_RWops *, Uint64, int, int, int);
 
 
 /* See if an image is contained in a data source */
@@ -240,6 +246,13 @@ static char * read_string (SDL_RWops * src) {
   return data;
 }
 
+static Uint64 read_offset (SDL_RWops * src, const xcf_header * h) {
+  Uint64 offset;  // starting with version 11, offsets are 64 bits
+  offset = (h->file_version >= 11) ? (Uint64)SDL_ReadBE32 (src) << 32 : 0;
+  offset |= SDL_ReadBE32 (src);
+  return offset;
+}
+
 
 static Uint32 Swap32 (Uint32 v) {
   return
@@ -249,13 +262,13 @@ static Uint32 Swap32 (Uint32 v) {
     |  ((v & 0xFF000000));
 }
 
-static void xcf_read_property (SDL_RWops * src, xcf_prop * prop) {
+static int xcf_read_property (SDL_RWops * src, xcf_prop * prop) {
   Uint32 len;
   prop->id = SDL_ReadBE32 (src);
   prop->length = SDL_ReadBE32 (src);
 
 #if DEBUG
-  printf ("%.8X: %s: %d\n", SDL_RWtell (src), prop->id < 25 ? prop_names [prop->id] : "unknown", prop->length);
+  printf ("%.8X: %s(%u): %u\n", SDL_RWtell (src), prop->id < 25 ? prop_names [prop->id] : "unknown", prop->id, prop->length);
 #endif
 
   switch (prop->id) {
@@ -286,8 +299,10 @@ static void xcf_read_property (SDL_RWops * src, xcf_prop * prop) {
     break;
   default:
     //    SDL_RWread (src, &prop->data, prop->length, 1);
-    SDL_RWseek (src, prop->length, RW_SEEK_CUR);
+    if (SDL_RWseek (src, prop->length, RW_SEEK_CUR) < 0)
+      return 0;  // ERROR
   }
+  return 1;  // OK
 }
 
 static void free_xcf_header (xcf_header * h) {
@@ -310,6 +325,18 @@ static xcf_header * read_xcf_header (SDL_RWops * src) {
   h->width       = SDL_ReadBE32 (src);
   h->height      = SDL_ReadBE32 (src);
   h->image_type  = SDL_ReadBE32 (src);
+  if (h->sign[9] == 'v' && h->sign[10] >= '0' && h->sign[10] <= '9' && h->sign[11] >= '0' && h->sign[11] <= '9' && h->sign[12] >= '0' && h->sign[12] <= '9')
+    h->file_version = (h->sign[10] - '0') * 100 + (h->sign[11] - '0') * 10 + (h->sign[12] - '0');
+  else
+    h->file_version = 0;
+#ifdef DEBUG
+  printf ("XCF signature : %.14s (version %u)\n", h->sign, h->file_version);
+  printf (" (%u,%u) type=%u\n", h->width, h->height, h->image_type);
+#endif
+  if (h->file_version >= 4)
+    h->precision = SDL_ReadBE32 (src);
+  else
+    h->precision = 150;
 
   h->properties = NULL;
   h->layer_file_offsets = NULL;
@@ -319,7 +346,10 @@ static xcf_header * read_xcf_header (SDL_RWops * src) {
 
   // Just read, don't save
   do {
-    xcf_read_property (src, &prop);
+    if (!xcf_read_property (src, &prop)) {
+      free_xcf_header (h);
+      return NULL;
+    }
     if (prop.id == PROP_COMPRESSION)
       h->compr = (xcf_compr_type)prop.data.compression;
     else if (prop.id == PROP_COLORMAP) {
@@ -351,7 +381,7 @@ static void free_xcf_layer (xcf_layer * l) {
   SDL_free (l);
 }
 
-static xcf_layer * read_xcf_layer (SDL_RWops * src) {
+static xcf_layer * read_xcf_layer (SDL_RWops * src, const xcf_header * h) {
   xcf_layer * l;
   xcf_prop    prop;
 
@@ -361,19 +391,27 @@ static xcf_layer * read_xcf_layer (SDL_RWops * src) {
   l->layer_type = SDL_ReadBE32 (src);
 
   l->name = read_string (src);
+#ifdef DEBUG
+  printf ("layer (%d,%d) type=%d '%s'\n", l->width, l->height, l->layer_type, l->name);
+#endif
 
   do {
-    xcf_read_property (src, &prop);
+    if (!xcf_read_property (src, &prop)) {
+      free_xcf_layer (l);
+      return NULL;
+    }
     if (prop.id == PROP_OFFSETS) {
       l->offset_x = prop.data.offset.x;
       l->offset_y = prop.data.offset.y;
     } else if (prop.id == PROP_VISIBLE) {
       l->visible = prop.data.visible ? 1 : 0;
+    } else if (prop.id == PROP_COLORMAP) {
+      SDL_free (prop.data.colormap.cmap);
     }
   } while (prop.id != PROP_END);
 
-  l->hierarchy_file_offset = SDL_ReadBE32 (src);
-  l->layer_mask_offset     = SDL_ReadBE32 (src);
+  l->hierarchy_file_offset = read_offset (src, h);
+  l->layer_mask_offset     = read_offset (src, h);
 
   return l;
 }
@@ -383,7 +421,7 @@ static void free_xcf_channel (xcf_channel * c) {
   SDL_free (c);
 }
 
-static xcf_channel * read_xcf_channel (SDL_RWops * src) {
+static xcf_channel * read_xcf_channel (SDL_RWops * src, const xcf_header * h) {
   xcf_channel * l;
   xcf_prop    prop;
 
@@ -392,10 +430,16 @@ static xcf_channel * read_xcf_channel (SDL_RWops * src) {
   l->height = SDL_ReadBE32 (src);
 
   l->name = read_string (src);
+#ifdef DEBUG
+  printf ("channel (%u,%u) '%s'\n", l->width, l->height, l->name);
+#endif
 
   l->selection = 0;
   do {
-    xcf_read_property (src, &prop);
+    if (!xcf_read_property (src, &prop)) {
+      free_xcf_channel (l);
+      return NULL;
+    }
     switch (prop.id) {
     case PROP_OPACITY:
       l->opacity = prop.data.opacity << 24;
@@ -416,7 +460,7 @@ static xcf_channel * read_xcf_channel (SDL_RWops * src) {
     }
   } while (prop.id != PROP_END);
 
-  l->hierarchy_file_offset = SDL_ReadBE32 (src);
+  l->hierarchy_file_offset = read_offset (src, h);
 
   return l;
 }
@@ -426,7 +470,7 @@ static void free_xcf_hierarchy (xcf_hierarchy * h) {
   SDL_free (h);
 }
 
-static xcf_hierarchy * read_xcf_hierarchy (SDL_RWops * src) {
+static xcf_hierarchy * read_xcf_hierarchy (SDL_RWops * src, const xcf_header * head) {
   xcf_hierarchy * h;
   int i;
 
@@ -438,8 +482,8 @@ static xcf_hierarchy * read_xcf_hierarchy (SDL_RWops * src) {
   h->level_file_offsets = NULL;
   i = 0;
   do {
-    h->level_file_offsets = (Uint32 *) SDL_realloc (h->level_file_offsets, sizeof (Uint32) * (i+1));
-    h->level_file_offsets [i] = SDL_ReadBE32 (src);
+    h->level_file_offsets = (Uint64 *) SDL_realloc (h->level_file_offsets, sizeof (*h->level_file_offsets) * (i+1));
+    h->level_file_offsets [i] = read_offset (src, head);
   } while (h->level_file_offsets [i++]);
 
   return h;
@@ -450,7 +494,7 @@ static void free_xcf_level (xcf_level * l) {
   SDL_free (l);
 }
 
-static xcf_level * read_xcf_level (SDL_RWops * src) {
+static xcf_level * read_xcf_level (SDL_RWops * src, const xcf_header * h) {
   xcf_level * l;
   int i;
 
@@ -461,8 +505,8 @@ static xcf_level * read_xcf_level (SDL_RWops * src) {
   l->tile_file_offsets = NULL;
   i = 0;
   do {
-    l->tile_file_offsets = (Uint32 *) SDL_realloc (l->tile_file_offsets, sizeof (Uint32) * (i+1));
-    l->tile_file_offsets [i] = SDL_ReadBE32 (src);
+    l->tile_file_offsets = (Uint64 *) SDL_realloc (l->tile_file_offsets, sizeof (*l->tile_file_offsets) * (i+1));
+    l->tile_file_offsets [i] = read_offset (src, h);
   } while (l->tile_file_offsets [i++]);
 
   return l;
@@ -472,27 +516,32 @@ static void free_xcf_tile (unsigned char * t) {
   SDL_free (t);
 }
 
-static unsigned char * load_xcf_tile_none (SDL_RWops * src, Uint32 len, int bpp, int x, int y) {
-  unsigned char * load;
+static unsigned char * load_xcf_tile_none (SDL_RWops * src, Uint64 len, int bpp, int x, int y) {
+  unsigned char * load = NULL;
 
-  load = (unsigned char *) SDL_malloc (len); // expect this is okay
-  SDL_RWread (src, load, len, 1);
+  if (len <= (Uint64)SDL_SIZE_MAX) {
+    load = (unsigned char *) SDL_malloc ((size_t)len); // expect this is okay
+  }
+  if (load != NULL)
+    SDL_RWread (src, load, (size_t)len, 1);
 
   return load;
 }
 
-static unsigned char * load_xcf_tile_rle (SDL_RWops * src, Uint32 len, int bpp, int x, int y) {
+static unsigned char * load_xcf_tile_rle (SDL_RWops * src, Uint64 len, int bpp, int x, int y) {
   unsigned char * load, * t, * data, * d;
-  Uint32 reallen;
   int i, size, count, j, length;
   unsigned char val;
 
-  if (len == 0) {  /* probably bogus data. */
+  if (len == 0 || len > (Uint64)SDL_SIZE_MAX) {  /* probably bogus data. */
     return NULL;
   }
 
-  t = load = (unsigned char *) SDL_malloc (len);
-  reallen = SDL_RWread (src, t, 1, len);
+  t = load = (unsigned char *) SDL_malloc ((size_t)len);
+  if (load == NULL)
+    return NULL;
+
+  SDL_RWread (src, t, 1, (size_t)len); /* reallen */
 
   data = (unsigned char *) SDL_calloc (1, x*y*bpp);
   for (i = 0; i < bpp; i++) {
@@ -590,14 +639,14 @@ do_layer_surface(SDL_Surface * surface, SDL_RWops * src, xcf_header * head, xcf_
     xcf_level      *level;
     unsigned char  *tile;
     Uint8          *p8;
-    Uint16         *p16;
     Uint32         *p;
     int            i, j;
     Uint32         x, y, tx, ty, ox, oy;
     Uint32         *row;
+    Uint64         length;
 
     SDL_RWseek(src, layer->hierarchy_file_offset, RW_SEEK_SET);
-    hierarchy = read_xcf_hierarchy(src);
+    hierarchy = read_xcf_hierarchy(src, head);
 
     if (hierarchy->bpp > 4) {  /* unsupported. */
         SDL_Log("Unknown Gimp image bpp (%u)\n", (unsigned int) hierarchy->bpp);
@@ -613,20 +662,23 @@ do_layer_surface(SDL_Surface * surface, SDL_RWops * src, xcf_header * head, xcf_
 
     level = NULL;
     for (i = 0; hierarchy->level_file_offsets[i]; i++) {
-        SDL_RWseek(src, hierarchy->level_file_offsets[i], RW_SEEK_SET);
-        level = read_xcf_level(src);
+        if (SDL_RWseek(src, hierarchy->level_file_offsets[i], RW_SEEK_SET) < 0)
+            break;
+        if (i > 0) // skip level except the 1st one, just like GIMP does
+            continue;
+        level = read_xcf_level(src, head);
 
         ty = tx = 0;
         for (j = 0; level->tile_file_offsets[j]; j++) {
             SDL_RWseek(src, level->tile_file_offsets[j], RW_SEEK_SET);
             ox = tx + 64 > level->width ? level->width % 64 : 64;
             oy = ty + 64 > level->height ? level->height % 64 : 64;
+            length = ox*oy*6;
 
-            if (level->tile_file_offsets[j + 1]) {
-                tile = load_tile(src, level->tile_file_offsets[j + 1] - level->tile_file_offsets[j], hierarchy->bpp, ox, oy);
-            } else {
-                tile = load_tile(src, ox * oy * 6, hierarchy->bpp, ox, oy);
+            if (level->tile_file_offsets[j + 1] > level->tile_file_offsets[j]) {
+                length = level->tile_file_offsets[j + 1] - level->tile_file_offsets[j];
             }
+            tile = load_tile(src, length, hierarchy->bpp, ox, oy);
 
             if (!tile) {
                 if (hierarchy) {
@@ -639,10 +691,9 @@ do_layer_surface(SDL_Surface * surface, SDL_RWops * src, xcf_header * head, xcf_
             }
 
             p8 = tile;
-            p16 = (Uint16 *) p8;
             p = (Uint32 *) p8;
             for (y = ty; y < ty + oy; y++) {
-                if ((ty >= surface->h) || ((tx+ox) > surface->w)) {
+                if ((y >= (Uint32)surface->h) || ((tx+ox) > (Uint32)surface->w)) {
                     break;
                 }
                 row = (Uint32 *) ((Uint8 *) surface->pixels + y * surface->pitch + tx * 4);
@@ -755,7 +806,7 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
   int chnls, i, offsets;
   Sint64 offset, fp;
 
-  unsigned char * (* load_tile) (SDL_RWops *, Uint32, int, int, int);
+  unsigned char * (* load_tile) (SDL_RWops *, Uint64, int, int, int);
 
   if (!src) {
     /* The error message has been set in SDL_RWFromFile */
@@ -795,7 +846,7 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
 
   offsets = 0;
 
-  while ((offset = SDL_ReadBE32 (src))) {
+  while ((offset = read_offset (src, head))) {
     head->layer_file_offsets = (Uint32 *) SDL_realloc (head->layer_file_offsets, sizeof (Uint32) * (offsets+1));
     head->layer_file_offsets [offsets] = (Uint32)offset;
     offsets++;
@@ -815,20 +866,22 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
     SDL_Rect rs, rd;
     SDL_RWseek (src, head->layer_file_offsets [i-1], RW_SEEK_SET);
 
-    layer = read_xcf_layer (src);
-    do_layer_surface (lays, src, head, layer, load_tile);
-    rs.x = 0;
-    rs.y = 0;
-    rs.w = layer->width;
-    rs.h = layer->height;
-    rd.x = layer->offset_x;
-    rd.y = layer->offset_y;
-    rd.w = layer->width;
-    rd.h = layer->height;
+    layer = read_xcf_layer (src, head);
+    if (layer != NULL) {
+      do_layer_surface (lays, src, head, layer, load_tile);
+      rs.x = 0;
+      rs.y = 0;
+      rs.w = layer->width;
+      rs.h = layer->height;
+      rd.x = layer->offset_x;
+      rd.y = layer->offset_y;
+      rd.w = layer->width;
+      rd.h = layer->height;
 
-    if (layer->visible)
-      SDL_BlitSurface (lays, &rs, surface, &rd);
-    free_xcf_layer (layer);
+      if (layer->visible)
+        SDL_BlitSurface (lays, &rs, surface, &rd);
+      free_xcf_layer (layer);
+    }
   }
 
   SDL_FreeSurface (lays);
@@ -838,11 +891,13 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
   // read channels
   channel = NULL;
   chnls   = 0;
-  while ((offset = SDL_ReadBE32 (src))) {
+  while ((offset = read_offset (src, head))) {
     channel = (xcf_channel **) SDL_realloc (channel, sizeof (xcf_channel *) * (chnls+1));
     fp = SDL_RWtell (src);
     SDL_RWseek (src, offset, RW_SEEK_SET);
-    channel [chnls++] = (read_xcf_channel (src));
+    channel [chnls] = (read_xcf_channel (src, head));
+    if (channel [chnls] != NULL)
+      chnls++;
     SDL_RWseek (src, fp, RW_SEEK_SET);
   }
 
@@ -884,6 +939,9 @@ done:
 }
 
 #else
+#if _MSC_VER >= 1300
+#pragma warning(disable : 4100) /* warning C4100: 'op' : unreferenced formal parameter */
+#endif
 
 /* See if an image is contained in a data source */
 int IMG_isXCF(SDL_RWops *src)
